@@ -1,21 +1,22 @@
 import typing
 import urllib.parse
+from datetime import datetime
 from timeit import default_timer as timer
 
 from bakplane.bakplane_pb2 import ResolveResourcePathResponse, ResolveResourceIntentResponse, ComparisonType
 from bakplane.extensions.base import WriteResponse, WriteRequest, ReadResponse, ReadAllRequest, ReadRequest, \
-    ReadRequestBuildingBlocks
-from bakplane.extensions.spark.utils import to_spark_save_mode, timestamp_to_date_string, datetime_to_date_string
+    ReadRequestBuildingBlocks, DataStoreHelper
+from bakplane.extensions.spark.utils import to_spark_save_mode, timestamp_to_date_string, datetime_to_date_string, \
+    timestamp_to_datetime
 from bakplane.utils import to_execution_statistics
 
-from datetime import datetime
-
 
 def __build_read_all_query(r: ReadAllRequest, res: ResolveResourcePathResponse) -> str:
     return f'select {", ".join(r.columns)} from {res.context["schema_name"]}.{res.context["table_name"]}'
 
 
-def __build_read_query(r: ReadRequest, res: ResolveResourceIntentResponse, dialect) -> str:
+def __build_read_query(r: ReadRequest, res: ResolveResourceIntentResponse, dialect,
+                       store_helper: DataStoreHelper) -> str:
     from sqlalchemy import (
         Table,
         MetaData,
@@ -56,6 +57,7 @@ def __build_read_query(r: ReadRequest, res: ResolveResourceIntentResponse, diale
                 m.c.get(clause.mapping_field_name) == a.c.get(clause.resource_field_name)
             )
 
+
     effective_start_dt = timestamp_to_date_string(res.effective_dating.end_dt.seconds)
     effective_end_dt = timestamp_to_date_string(res.effective_dating.end_dt.seconds)
 
@@ -63,48 +65,58 @@ def __build_read_query(r: ReadRequest, res: ResolveResourceIntentResponse, diale
         mapping_table=m,
         asset_table=a,
         selectable=m.join(a, and_(*clauses)),
-        effective_start_dt=effective_start_dt,
-        effective_end_dt=effective_end_dt,
+        effective_start_dt=timestamp_to_datetime(res.effective_dating.start_dt.seconds),
+        effective_end_dt=timestamp_to_datetime(res.effective_dating.end_dt.seconds),
         knowledge_start_dt=r.knowledge_start_dt,
-        knowledge_end_dt=r.knowledge_end_dt
+        knowledge_end_dt=r.knowledge_end_dt,
+        constraints=[]
     )
 
+    if r.columns:
+        blocks.columns = [a.c.get(x) for x in r.columns if x in a.c]
+    else:
+        blocks.columns = [a.c.get(x) for x in res.columns]
+
     if r.query_builder:
         res = r.query_builder(blocks)
     else:
 
-        constraints = []
+        blocks.constraints = []
 
         if r.additional_constraints is not None:
-            constraints.append(r.additional_constraints)
+            blocks.constraints.append(r.additional_constraints)
 
         if r.effective_dating_hint is None:
-            constraints.append(a.c.effective_start_dt <= m.c.effective_end_dt)
-            constraints.append(a.c.effective_end_dt >= m.c.effective_start_dt)
-            constraints.append(a.c.effective_start_dt <= timestamp_to_date_string(res.effective_dating.end_dt.seconds))
-            constraints.append(a.c.effective_end_dt >= timestamp_to_date_string(res.effective_dating.start_dt.seconds))
+            blocks.constraints.append(a.c.effective_start_dt <= m.c.effective_end_dt)
+            blocks.constraints.append(a.c.effective_end_dt >= m.c.effective_start_dt)
+            blocks.constraints.append(a.c.effective_start_dt <= timestamp_to_date_string(res.effective_dating.end_dt.seconds))
+            blocks.constraints.append(a.c.effective_end_dt >= timestamp_to_date_string(res.effective_dating.start_dt.seconds))
         else:
             if isinstance(r.effective_dating_hint, tuple):
                 s = datetime_to_date_string(typing.cast(datetime, r.effective_dating_hint[0]))
                 e = datetime_to_date_string(typing.cast(datetime, r.effective_dating_hint[1]))
 
-                constraints.append(a.c.effective_start_dt <= e)
-                constraints.append(a.c.effective_end_dt >= s)
+                blocks.constraints.append(a.c.effective_start_dt <= e)
+                blocks.constraints.append(a.c.effective_end_dt >= s)
             elif isinstance(r.effective_dating_hint, str):
                 col = typing.cast(str, r.effective_dating_hint)
 
                 s = col + '_start_dt'
                 e = col + '_end_dt'
 
-                constraints.append(a.c.effective_start_dt <= m.c.get(e))
-                constraints.append(a.c.effective_end_dt >= m.c.get(s))
+                blocks.constraints.append(a.c.effective_start_dt <= m.c.get(e))
+                blocks.constraints.append(a.c.effective_end_dt >= m.c.get(s))
             else:
                 raise RuntimeError('Invalid type for effective_dating_hint.')
 
+        if r.modifiers:
+            for m in r.modifiers:
+                blocks = m.execute(blocks, store_helper)
+
         res = (
-            select([a.c.get(x) for x in r.columns])
-                .select_from(m.join(a, and_(*clauses)))
-                .where(and_(*constraints))
+            select(blocks.columns)
+                .select_from(blocks.selectable)
+                .where(and_(*blocks.constraints))
         )
 
     query = str(
@@ -172,12 +184,12 @@ def read_all_fn(ctx: typing.Any, r: ReadAllRequest, res: ResolveResourcePathResp
 
 
 def read_fn(ctx: typing.Any, r: ReadRequest, res: ResolveResourceIntentResponse, dialect, url: str,
-            driver: str) -> ReadResponse:
+            driver: str, store_helper: DataStoreHelper) -> ReadResponse:
     return __read_query(
         spark=ctx,
         url=url,
         driver=driver,
-        query=__build_read_query(r, res, dialect),
+        query=__build_read_query(r, res, dialect, store_helper),
         user=res.context['username'],
         password=res.context['password']
     )
